"""
hubspot contact lists api
"""
from hubspot3 import logging_helper
from hubspot3.base import BaseClient


CONTACT_LISTS_API_VERSION = "1"


class ContactListsClient(BaseClient):
    """
    The hubspot3 Contact Lists client uses the _make_request method to call the API for data.
    It returns a python object translated from the json returned
    """

    def __init__(self, *args, **kwargs):
        super(ContactListsClient, self).__init__(*args, **kwargs)
        self.log = logging_helper.get_log("hubspot3.contact_lists")

    def _get_path(self, subpath):
        return "contacts/v{}/{}".format(
            self.options.get("version") or CONTACT_LISTS_API_VERSION, subpath
        )

    def get_contact_lists(self, **options):
        """Returns all of the contact lists"""
        return self._call("lists", method="GET", **options)

    def add_contact_to_a_list(self, list_id, vids, data=None, **options):
        """Adds a list of contact vids to the specified list."""
        data = data or {}
        data["vids"] = vids
        return self._call(
            "lists/{list_id}/add".format(list_id=list_id),
            data=data,
            method="POST",
            **options
        )

    def create_a_contact_list(
        self, list_name, portal_id, dynamic=True, data=None, **options
    ):
        """Creates a contact list with given list_name on the given portal_id."""
        data = data or {}
        data["name"] = list_name
        data["portal_id"] = portal_id
        data["dynamic"] = dynamic
        return self._call("lists", data=data, method="POST", **options)

    def delete_a_contact_list(self, list_id, **options):
        """Deletes the contact list by list_id."""
        return self._call(
            "lists/{list_id}".format(list_id=list_id), method="DELETE", **options
        )



"""
base hubspot client class
"""
import urllib.request
import urllib.parse
import urllib.error
import http.client
import json
import logging
import time
import traceback
import zlib
from hubspot3 import utils
from hubspot3.utils import force_utf8
from hubspot3.error import (
    HubspotBadConfig,
    HubspotBadRequest,
    HubspotConflict,
    HubspotError,
    HubspotNoConfig,
    HubspotNotFound,
    HubspotRateLimited,
    HubspotServerError,
    HubspotTimeout,
    HubspotUnauthorized,
)
from typing import Union


class BaseClient(object):
    """Base abstract object for interacting with the HubSpot APIs"""

    # Controls how long we sleep for during retries, overridden by unittests
    # so tests run faster
    sleep_multiplier = 1

    def __init__(
        self,
        api_key=None,
        access_token=None,
        refresh_token=None,
        client_id=None,
        timeout=10,
        mixins=None,
        api_base="api.hubapi.com",
        debug=False,
        disable_auth=False,
        **extra_options
    ):
        super(BaseClient, self).__init__()
        # reverse so that the first one in the list because the first parent
        if not mixins:
            mixins = []
        mixins.reverse()
        for mixin_class in mixins:
            if mixin_class not in self.__class__.__bases__:
                self.__class__.__bases__ = (mixin_class,) + self.__class__.__bases__

        self.api_key = api_key or extra_options.get("api_key")
        self.access_token = access_token or extra_options.get("access_token")
        self.refresh_token = refresh_token or extra_options.get("refresh_token")
        self.client_id = client_id or extra_options.get("client_id")
        self.log = utils.get_log("hubspot3")
        if not disable_auth:
            if self.api_key and self.access_token:
                raise HubspotBadConfig("Cannot use both api_key and access_token.")
            if not (self.api_key or self.access_token or self.refresh_token):
                raise HubspotNoConfig("Missing required credentials.")
        self.options = {
            "api_base": api_base,
            "debug": debug,
            "disable_auth": disable_auth,
            "timeout": timeout,
        }
        self.options.update(extra_options)
        self._prepare_connection_type()

    def _prepare_connection_type(self):
        connection_types = {
            "http": http.client.HTTPConnection,
            "https": http.client.HTTPSConnection,
        }
        parts = self.options["api_base"].split("://")
        protocol = (parts[0:-1] + ["https"])[0]
        self.options["connection_type"] = connection_types[protocol]
        self.options["protocol"] = protocol
        self.options["api_base"] = parts[-1]

    def _get_path(self, subpath):
        """get the full api url for the given subpath on this client"""
        return subpath

    def _prepare_request_auth(self, subpath, params, data, opts):
        if self.api_key:
            params["hapikey"] = params.get("hapikey") or self.api_key

    def _prepare_request(self, subpath, params, data, opts, doseq=False, query=""):
        params = params or {}
        self._prepare_request_auth(subpath, params, data, opts)

        if opts.get("hub_id") or opts.get("portal_id"):
            params["portalId"] = opts.get("hub_id") or opts.get("portal_id")
        if query is None:
            query = ""
        if query and query.startswith("?"):
            query = query[1:]
        if query and not query.startswith("&"):
            query = "&" + query
        url = opts.get("url") or "/{}?{}{}".format(
            self._get_path(subpath), urllib.parse.urlencode(params, doseq), query
        )
        headers = opts.get("headers") or {}
        headers.update(
            {
                "Accept-Encoding": "gzip",
                "Content-Type": opts.get("content_type") or "application/json",
            }
        )
        if self.access_token:
            headers.update({"Authorization": "Bearer {}".format(self.access_token)})

        if data and headers["Content-Type"] == "application/json":
            data = json.dumps(data)

        return url, headers, data

    def _create_request(self, conn, method, url, headers, data):
        conn.request(method, url, data, headers)
        params = {"method": method, "url": url, "data": data, "headers": headers, "host": conn.host}
        params["timeout"] = conn.timeout
        return params

    def _gunzip_body(self, body):
        return zlib.decompress(body)

    def _process_body(self, data, gzipped):
        if gzipped:
            return force_utf8(self._gunzip_body(data))
        return data

    def _execute_request_raw(self, conn, request):
        try:
            result = conn.getresponse()
        except Exception:
            raise HubspotTimeout(None, request, traceback.format_exc())

        encoding = [i[1] for i in result.getheaders() if i[0] == "content-encoding"]
        possibly_encoded = result.read()
        try:
            possibly_encoded = zlib.decompress(possibly_encoded, 16 + zlib.MAX_WBITS)
        except Exception:
            pass
        result.body = self._process_body(possibly_encoded, len(encoding) and encoding[0] == "gzip")

        conn.close()
        if result.status in (404, 410):
            raise HubspotNotFound(result, request)
        elif result.status == 401:
            raise HubspotUnauthorized(result, request)
        elif result.status == 409:
            raise HubspotConflict(result, request)
        elif result.status == 429:
            raise HubspotRateLimited(result, request)
        elif result.status >= 400 and result.status < 500 or result.status == 501:
            raise HubspotBadRequest(result, request)
        elif result.status >= 500:
            raise HubspotServerError(result, request)

        return result

    def _execute_request(self, conn, request):
        result = self._execute_request_raw(conn, request)
        return result.body

    def _digest_result(self, data):
        if data:
            if isinstance(data, str):
                try:
                    data = json.loads(data)
                except ValueError:
                    pass
            elif isinstance(data, bytes):
                data = json.loads(utils.force_utf8(data))

        return data

    def _prepare_request_retry(self, method, url, headers, data):
        pass

    def _call_raw(
        self,
        subpath,
        params=None,
        method="GET",
        data=None,
        doseq=False,
        query="",
        retried=False,
        **options
    ):
        opts = self.options.copy()
        opts.update(options)

        debug = opts.get("debug")

        url, headers, data = self._prepare_request(
            subpath, params, data, opts, doseq=doseq, query=query
        )

        if debug:
            print(
                json.dumps({"url": url, "headers": headers, "data": data}, sort_keys=True, indent=2)
            )

        kwargs = {}
        kwargs["timeout"] = opts["timeout"]

        num_retries = opts.get("number_retries", 2)

        # Never retry a POST, PUT, or DELETE unless explicitly told to
        if method != "GET" and not opts.get("retry_on_post"):
            num_retries = 0
        if num_retries > 6:
            num_retries = 6

        emergency_brake = 10
        try_count = 0
        while True:
            emergency_brake -= 1
            # avoid getting burned by any mistakes in While loop logic
            if emergency_brake < 1:
                break
            try:
                try_count += 1
                connection = opts["connection_type"](opts["api_base"], **kwargs)
                request_info = self._create_request(connection, method, url, headers, data)
                result = self._execute_request_raw(connection, request_info)
                break
            except HubspotUnauthorized:
                self.log.warning("401 Unauthorized response to API request.")
                if self.access_token and self.refresh_token and self.client_id and not retried:
                    self.log.info("Refreshing access token")
                    try:
                        token_response = utils.refresh_access_token(
                            self.refresh_token, self.client_id
                        )
                        decoded = json.loads(token_response)
                        self.access_token = decoded["access_token"]
                        self.log.info("Retrying with new token {}".format(self.access_token))
                    except Exception as exception:
                        self.log.error("Unable to refresh access_token: {}".format(exception))
                        raise
                    return self._call_raw(
                        subpath,
                        params=params,
                        method=method,
                        data=data,
                        doseq=doseq,
                        query=query,
                        retried=True,
                        **options
                    )
                else:
                    if self.access_token and self.refresh_token and self.client_id and retried:
                        self.log.error(
                            "Refreshed token, but request still was not authorized. "
                            " You may need to grant additional permissions."
                        )
                    elif self.access_token and not self.refresh_token:
                        self.log.error(
                            "In order to enable automated refreshing of your access "
                            "token, please provide a refresh token as well."
                        )
                    elif self.access_token and not self.client_id:
                        self.log.error(
                            "In order to enable automated refreshing of your access "
                            "token, please provide a client_id in addition to a refresh token."
                        )
                    raise
            except HubspotError as exception:
                if try_count > num_retries:
                    logging.warning("Too many retries for {}".format(url))
                    raise
                # Don't retry errors from 300 to 499
                if (
                    exception.result
                    and exception.result.status >= 300
                    and exception.result.status < 500
                ):
                    raise
                self._prepare_request_retry(method, url, headers, data)
                self.log.warning("HubspotError {} calling {}, retrying".format(exception, url))
            # exponential back off
            # wait 0 seconds, 1 second, 3 seconds, 7 seconds, 15 seconds, etc
            time.sleep((pow(2, try_count - 1) - 1) * self.sleep_multiplier)
        return result

    def _call(
        self,
        subpath: str,
        params: dict = None,
        method: str = "GET",
        data: Union[str, dict, list] = None,
        doseq: bool = False,
        query: str = "",
        raw: bool = False,
        **options
    ):
        result = self._call_raw(
            subpath,
            params=params,
            method=method,
            data=data,
            doseq=doseq,
            query=query,
            retried=False,
            **options
        )
        return result if raw else self._digest_result(result.body)



"""
hubspot leads api
"""
import time
from hubspot3.base import BaseClient
from hubspot3 import logging_helper


LEADS_API_VERSION = "1"


def list_to_snake_dict(list_):
    dictionary = {}
    for item in list_:
        dictionary[item] = item
        if item.lower() != item:
            python_variant = item[0].lower() + "".join(
                [c if c.lower() == c else "_{}".format(c.lower()) for c in item[1:]]
            )
            dictionary[python_variant] = item
    return dictionary


SORT_OPTIONS = [
    "firstName",
    "lastName",
    "email",
    "address",
    "phone",
    "insertedAt",
    "fce.convertDate",
    "lce.convertDate",
    "lastModifiedAt",
    "closedAt",
]
SORT_OPTIONS_DICT = list_to_snake_dict(SORT_OPTIONS)
TIME_PIVOT_OPTIONS = [
    "insertedAt",
    "firstConvertedAt",
    "lastConvertedAt",
    "lastModifiedAt",
    "closedAt",
]
TIME_PIVOT_OPTIONS_DICT = list_to_snake_dict(TIME_PIVOT_OPTIONS)
SEARCH_OPTIONS = [
    "search",
    "sort",
    "dir",
    "max",
    "offset",
    "startTime",
    "stopTime",
    "timePivot",
    "excludeConversionEvents",
    "emailOptOut",
    "eligibleForEmail",
    "bounced",
    "isNotImported",
]
SEARCH_OPTIONS_DICT = list_to_snake_dict(SEARCH_OPTIONS)
BOOLEAN_SEARCH_OPTIONS = set(
    [
        "excludeConversionEvents",
        "emailOptOut",
        "eligibleForEmail",
        "bounced",
        "isNotImported",
    ]
)

MAX_BATCH = 100


class LeadsClient(BaseClient):
    """
    The hubspot3 Leads client uses the _make_request method to call the API for data.
    It returns a python object translated from the json returned
    """

    def __init__(self, *args, **kwargs):
        super(LeadsClient, self).__init__(*args, **kwargs)
        self.log = logging_helper.get_log("hubspot3.leads")

    def camelcase_search_options(self, options):
        """change all underscored variants back to what the API is expecting"""
        new_options = {}
        for key in options:
            value = options[key]
            new_key = SEARCH_OPTIONS_DICT.get(key, key)
            if new_key == "sort":
                value = SORT_OPTIONS_DICT.get(value, value)
            elif new_key == "timePivot":
                value = TIME_PIVOT_OPTIONS_DICT.get(value, value)
            elif new_key in BOOLEAN_SEARCH_OPTIONS:
                value = str(value).lower()
            new_options[new_key] = value
        return new_options

    def _get_path(self, subpath):
        """get the full api url for the given subpath on this client"""
        return "leads/v{}/{}".format(
            self.options.get("version") or LEADS_API_VERSION, subpath
        )

    def get_lead(self, guid, **options):
        return self.get_leads(guid, **options)[0]

    def get_leads(self, *guids, **options):
        """Supports all the search parameters in the API as well as python underscored variants"""
        original_options = options
        options = self.camelcase_search_options(options.copy())
        params = {}
        for i, guid in enumerate(guids):
            params["guids[{}]".format(i)] = guid
        for k in list(options.keys()):
            if k in SEARCH_OPTIONS:
                params[k] = options[k]
                del options[k]
        leads = self._call("list/", params, **options)
        self.log.info(
            "retrieved {} leads through API ( {}options={} )".format(
                len(leads), guids and "guids={}, ".format(guids or ""), original_options
            )
        )
        return leads

    def retrieve_lead(self, *guid, **options):
        cur_guid = guid or ""
        params = {}
        for key in options:
            params[key] = options[key]
        # Set guid to -1 as default for not finding a user
        lead = {"guid": "-1"}
        # wrap lead call so that it doesn't error out when not finding a lead
        try:
            lead = self._call("lead/{}".format(cur_guid), params, **options)
        except Exception:
            # no lead here
            pass
        return lead

    def update_lead(self, guid, update_data=None, **options):
        update_data = update_data or {}
        update_data["guid"] = guid
        return self._call(
            "lead/{}/".format(guid), data=update_data, method="PUT", **options
        )

    def get_webhook(self, **options):  # WTF are these 2 methods for?
        return self._call("callback-url", **options)

    def register_webhook(self, url, **options):
        return self._call(
            "callback-url",
            params={"url": url},
            data={"url": url},
            method="POST",
            **options
        )

    def close_lead(self, guid, close_time=None, **options):
        return self.update_lead(
            guid, {"closedAt": close_time or int(time.time() * 1000)}, **options
        )

    def open_lead(self, guid, **options):
        self.update_lead(guid, {"closedAt": ""}, **options)



"""
logging helper function
"""
import logging


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def get_log(name):
    logger = logging.getLogger(name)
    logger.addHandler(NullHandler())
    return logger



"""
base utils for the hubspot3 library
"""
import requests
import logging
from hubspot3.globals import BASE_URL


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def get_log(name):
    logger = logging.getLogger(name)
    logger.addHandler(NullHandler())
    return logger


def auth_checker(access_token: str) -> int:
    """Do a simple api request using the access token"""
    url = "{}/contacts/v1/lists/all/contacts/all?count=1&offset=0&access_token={}".format(
        BASE_URL, access_token
    )
    result = requests.get(url)
    return result.status_code


def refresh_access_token(refresh_token: str, client_id: str) -> str:
    """Refreshes an OAuth access token"""
    payload = "refresh_token={}&client_id={}&grant_type=refresh_token".format(
        refresh_token, client_id
    )
    url = "{}/oauth/v1/token".format(BASE_URL)
    result = requests.post(url, data=payload)
    return result.text


def force_utf8(raw):
    """Will force the string to convert to valid utf8"""
    string = raw
    try:
        string = string.decode("utf-8", "ignore")
    except Exception:
        pass
    return string


def prettify(obj_with_props, id_key):
    prettified = {
        prop: obj_with_props["properties"][prop]["value"]
        for prop in obj_with_props["properties"]
    }
    prettified["id"] = obj_with_props[id_key]
    try:
        prettified.update(
            {
                assoc: obj_with_props["associations"][assoc]
                for assoc in obj_with_props["associations"]
            }
        )
    except KeyError:
        pass

    return prettified



"""
hubspot contacts api
"""
from hubspot3 import logging_helper
from hubspot3.base import BaseClient
from hubspot3.utils import prettify
from typing import Union


CONTACTS_API_VERSION = "1"


class ContactsClient(BaseClient):
    """
    The hubspot3 Contacts client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    class Recency:
        """recency type enum"""

        CREATED = "created"
        MODIFIED = "modified"

    def __init__(self, *args, **kwargs):
        """initialize a contacts client"""
        super(ContactsClient, self).__init__(*args, **kwargs)
        self.log = logging_helper.get_log("hubspot3.contacts")

    def _get_path(self, subpath):
        return "contacts/v{}/{}".format(CONTACTS_API_VERSION, subpath)

    def create_or_update_a_contact(self, email, data=None, **options):
        """Creates or Updates a client with the supplied data."""
        data = data or {}
        return self._call(
            "contact/createOrUpdate/email/{email}".format(email=email),
            data=data,
            method="POST",
            **options
        )

    def get_contact_by_email(self, email, **options):
        """Gets contact specified by email address."""
        return self._call(
            "contact/email/{email}/profile".format(email=email), method="GET", **options
        )

    def get_contact_by_id(self, contact_id: str, **options):
        """Gets contact specified by ID"""
        return self._call(
            "contact/vid/{}/profile".format(contact_id), method="GET", **options
        )

    def update_a_contact(self, contact_id, data=None, **options):
        """Updates the contact by contact_id with the given data."""
        data = data or {}
        return self._call(
            "contact/vid/{contact_id}/profile".format(contact_id=contact_id),
            data=data,
            method="POST",
            **options
        )

    def delete_a_contact(self, contact_id: str, **options):
        """Deletes a contact by contact_id."""
        return self._call(
            "contact/vid/{contact_id}".format(contact_id=contact_id),
            method="DELETE",
            **options
        )

    def create(self, data=None, **options):
        """create a contact"""
        data = data or {}
        return self._call("contact", data=data, method="POST", **options)

    def update(self, contact_id: str, data=None, **options):
        """update the given vid with the given data"""
        if not data:
            data = {}

        return self._call(
            "contact/vid/{}/profile".format(contact_id),
            data=data,
            method="POST",
            **options
        )

    def get_batch(self, ids, extra_properties: Union[list, str] = None):
        """given a batch of vids, get more of their info"""
        # default properties to fetch
        properties = [
            "email",
            "firstname",
            "lastname",
            "company",
            "website",
            "phone",
            "address",
            "city",
            "state",
            "zip",
            "associatedcompanyid",
        ]

        # append extras if they exist
        if extra_properties:
            if isinstance(extra_properties, list):
                properties += extra_properties
            if isinstance(extra_properties, str):
                properties.append(extra_properties)

        batch = self._call(
            "contact/vids/batch",
            method="GET",
            doseq=True,
            params={"vid": ids, "property": properties},
        )
        # It returns a dict with IDs as keys
        return [prettify(batch[contact], id_key="vid") for contact in batch]

    def get_all(
        self, extra_properties: Union[list, str] = None, limit: int = -1, **options
    ) -> list:
        """
        get all contacts in hubspot, fetching additional properties if passed in
        Can't get phone number from a get-all, so we just grab IDs and
        then have to make ANOTHER call in batches
        :see: https://developers.hubspot.com/docs/methods/contacts/get_contacts
        """
        finished = False
        output = []  # type: list
        offset = 0
        query_limit = 100  # Max value according to docs
        limited = limit > 0
        if limited and limit < query_limit:
            query_limit = limit
        while not finished:
            batch = self._call(
                "lists/all/contacts/all",
                method="GET",
                params={"count": query_limit, "vidOffset": offset},
                **options
            )
            output.extend(
                self.get_batch(
                    [contact["vid"] for contact in batch["contacts"]],
                    extra_properties=extra_properties,
                )
            )
            finished = not batch["has-more"] or (limited and len(output) >= limit)
            offset = batch["vid-offset"]

        return output if not limited else output[:limit]

    def _get_recent(
        self,
        recency_type: str,
        limit: int = 100,
        vid_offset: int = 0,
        time_offset: int = 0,
        **options
    ):
        """
        returns a list of either recently created or recently modified/created contacts
        """
        finished = False
        output = []
        query_limit = 100  # max according to the docs
        recency_string = (
            "all"
            if recency_type == ContactsClient.Recency.CREATED
            else "recently_updated"
        )
        limited = limit > 0
        if limited and limit < query_limit:
            query_limit = limit

        while not finished:
            params = {"count": query_limit}
            if vid_offset and time_offset:
                params["vidOffset"] = vid_offset
                params["timeOffset"] = time_offset
            batch = self._call(
                "lists/{}/contacts/recent".format(recency_string),
                method="GET",
                params=params,
                doseq=True,
                **options
            )
            output.extend([contact for contact in batch["contacts"]])
            finished = not batch["has-more"] or len(output) >= limit
            vid_offset = batch["vid-offset"]
            time_offset = batch["time-offset"]

        return output[:limit]

    def get_recently_created(self, limit: int = 100):
        """
        get recently created contacts
        :see: https://developers.hubspot.com/docs/methods/contacts/get_recently_created_contacts
        """
        return self._get_recent(ContactsClient.Recency.CREATED, limit=limit)

    def get_recently_modified(self, limit: int = 100):
        """
        get recently modified and created contacts
        :see: https://developers.hubspot.com/docs/methods/contacts/get_recently_updated_contacts
        """
        return self._get_recent(ContactsClient.Recency.MODIFIED, limit=limit)



"""
hubspot owners api
"""
from hubspot3.base import BaseClient


OWNERS_API_VERSION = "v2"


class OwnersClient(BaseClient):
    """
    hubspot3 Owners client
    :see: https://developers.hubspot.com/docs/methods/owners/owners_overview
    """

    def _get_path(self, subpath):
        """get the full api url for the given subpath on this client"""
        return "owners/{}/owners".format(OWNERS_API_VERSION)

    def get_owners(self, **options):
        """Only returns the list of owners, does not include additional metadata"""
        return self._call("owners", **options)

    def get_owner_name_by_id(self, owner_id: str, **options) -> str:
        """given an id of an owner, return their name"""
        owner_name = "value_missing"
        owners = self.get_owners()
        for owner in owners:
            if int(owner["ownerId"]) == int(owner_id):
                owner_name = "{} {}".format(owner["firstName"], owner["lastName"])
        return owner_name

    def get_owner_email_by_id(self, owner_id: str, **options) -> str:
        """given an id of an owner, return their email"""
        owner_email = "value_missing"
        owners = self.get_owners()
        for owner in owners:
            if int(owner["ownerId"]) == int(owner_id):
                owner_email = owner["email"]
        return owner_email



"""
hubspot crm_associations api
"""
from hubspot3 import logging_helper
from hubspot3.base import BaseClient
from enum import Enum


ASSOCIATIONS_API_VERSION = "1"


class Definitions(Enum):
    CONTACT_TO_COMPANY = 1
    COMPANY_TO_CONTACT = 2
    DEAL_TO_CONTACT = 3
    CONTACT_TO_DEAL = 4
    DEAL_TO_COMPANY = 5
    COMPANY_TO_DEAL = 6
    COMPANY_TO_ENGAGEMENT = 7
    ENGAGEMENT_TO_COMPANY = 8
    CONTACT_TO_ENGAGEMENT = 9
    ENGAGEMENT_TO_CONTACT = 10
    DEAL_TO_ENGAGEMENT = 11
    ENGAGEMENT_TO_DEAL = 12
    PARENT_COMPANY_TO_CHILD_COMPANY = 13
    CHILD_COMPANY_TO_PARENT_COMPANY = 14
    CONTACT_TO_TICKET = 15
    TICKET_TO_CONTACT = 16
    TICKET_TO_ENGAGEMENT = 17
    ENGAGEMENT_TO_TICKET = 18
    DEAL_TO_LINE_ITEM = 19
    LINE_ITEM_TO_DEAL = 20
    COMPANY_TO_TICKET = 25
    TICKET_TO_COMPANY = 26
    DEAL_TO_TICKET = 27
    TICKET_TO_DEAL = 28


class CRMAssociationsClient(BaseClient):
    """
    Associations extension for Associations API endpoint
    :see: https://developers.hubspot.com/docs/methods/crm-associations/crm-associations-overview
    """

    def __init__(self, *args, **kwargs):
        super(CRMAssociationsClient, self).__init__(*args, **kwargs)
        self.log = logging_helper.get_log("hubspot3.crm_associations")

    def _get_path(self, subpath):
        return "crm-associations/v{}/{}".format(
            self.options.get("version") or ASSOCIATIONS_API_VERSION, subpath
        )

    def get(self, object_id, definition: Definitions):
        """
        get all associations for the defined object
        :param object_id: Object ID for the object you're looking up
        :param definition: Definition ID for the objects you're looking for associations of
        """
        finished = False
        output = []
        offset = 0
        query_limit = 100  # Max value according to docs

        while not finished:
            batch = self._call(
                "associations/{}/{}/{}".format(
                    object_id, "HUBSPOT_DEFINED", definition.value
                ),
                method="GET",
                params={"limit": query_limit, "offset": offset},
            )
            output.extend([id_ for id_ in batch["results"]])
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output

    def get_all(self, object_id, definition: Definitions, **options):
        """
        get all crm associations
        :param object_id: Object ID for the object you're looking up
        :param definition: Definition ID for the objects you're looking for associations of
        """
        finished = False
        output = []
        offset = 0
        query_limit = 100  # Max value according to docs

        while not finished:
            batch = self._call(
                "associations/{}/{}/{}".format(
                    object_id, "HUBSPOT_DEFINED", definition.value
                ),
                method="GET",
                params={"limit": query_limit, "offset": offset},
            )
            output.extend([id_ for id_ in batch["results"]])
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output

    def create(self, from_object, to_object, definition: Definitions, **options):
        """
        create a hubspot association
        :param from_object: ID of object to relate
        :param to_object: ID of object to relate to
        :param definition: Definition ID for the objects you're looking for associations of
        """
        return self._call(
            "associations",
            method="PUT",
            data={
                "fromObjectId": from_object,
                "toObjectId": to_object,
                "category": "HUBSPOT_DEFINED",
                "definitionId": definition.value,
            },
            **options
        )

    def delete(self, from_object, to_object, definition: Definitions, **options):
        """
        delete a hubspot association
        :param from_object: ID of object to relate
        :param to_object: ID of object to relate to
        :param definition: Definition ID for the objects you're looking for associations of
        """
        return self._call(
            "associations/delete",
            method="PUT",
            data={
                "fromObjectId": from_object,
                "toObjectId": to_object,
                "category": "HUBSPOT_DEFINED",
                "definitionId": definition.value,
            },
            **options
        )



"""
hubspot prospects client
"""
from hubspot3.base import BaseClient


PROSPECTS_API_VERSION = "v1"


class ProspectsClient(BaseClient):
    """
    Python client for the HubSpot Prospects API.
    This client provides convenience methods for the HubSpot Prospects API.
    It is a work in progress, and contributions are welcome.
    """

    def _get_path(self, subpath):
        """get the full api url for the given subpath on this client"""
        return "prospects/{}/{}".format(PROSPECTS_API_VERSION, subpath)

    def get_prospects(self, offset=None, orgoffset=None, limit=None):
        """
        Return the prospects for the current API key.

        Optionally start the result list at the given offset.

        Each member of the return list is a prospect element containing
        organizational information such as name and location.
        """
        params = {}
        if limit:
            params["count"] = limit

        if offset:
            params["timeOffset"] = offset
            params["orgOffset"] = orgoffset

        return self._call("timeline", params)

    def get_company(self, company_slug):
        """Return the specific named organization for the given API key, if we find a match."""
        return self._call("timeline/{}".format(company_slug))

    def get_options_for_query(self, query):
        """This method allows for discovery of prospects with partial names."""
        return self._call("typeahead/", {"q": query})

    def search_prospects(self, search_type, query, offset=None, orgoffset=None):
        """
        Supports doing a search for prospects by city, reion, or country.

        search_type should be one of 'city' 'region' 'country'.

        This method is intended to be called with one of the outputs from the
        get_options_for_query method above.
        """
        params = {"q": query}
        if offset and orgoffset:
            params["orgOffset"] = orgoffset
            params["timeOffset"] = offset

        return self._call("search/{}".format(search_type), params)

    def get_hidden_prospects(self):
        """Return the list of prospects hidden by the customer (or this API), if any."""
        return self._call("filters")

    def hide_prospect(self, company_name):
        """Hides the given prospect from the user interface."""
        return self._call(
            "filters",
            data=("organization={}".format(company_name)),
            method="POST",
            content_type="application/x-www-form-urlencoded",
        )

    def unhide_prospect(self, company_name):
        """Un-hides, i.e. displays, the given prospect in the user interface."""
        return self._call(
            "filters", data={"organization": company_name}, method="DELETE"
        )
